@@ -35,6 +35,7 @@ QPDFObjectHandle::ParserCallbacks::terminateParsing()
 
 QPDFObjectHandle::QPDFObjectHandle() :
     initialized(false),
+    qpdf(0),
     objid(0),
     generation(0),
     reserved(false)
@@ -535,6 +536,12 @@ QPDFObjectHandle::replaceStreamData(PointerHolder<StreamDataProvider> provider,
 	provider, filter, decode_parms);
 }
 
+QPDFObjGen
+QPDFObjectHandle::getObjGen() const
+{
+    return QPDFObjGen(this->objid, this->generation);
+}
+
 int
 QPDFObjectHandle::getObjectID() const
 {
