@@ -2,6 +2,7 @@
 
 #include <string.h>
 #include <stdio.h>
+#include <qpdf/QUtil.hh>
 
 QPDF_Name::QPDF_Name(std::string const& name) :
     name(name)
@@ -16,7 +17,6 @@ std::string
 QPDF_Name::normalizeName(std::string const& name)
 {
     std::string result;
-    char num[4];
     result += name[0];
     for (unsigned int i = 1; i < name.length(); ++i)
     {
@@ -24,8 +24,7 @@ QPDF_Name::normalizeName(std::string const& name)
 	// Don't use locale/ctype here; follow PDF spec guidelines.
 	if (strchr("#()<>[]{}/%", ch) || (ch < 33) || (ch > 126))
 	{
-	    sprintf(num, "#%02x", (unsigned char) ch);
-	    result += num;
+            result += "#" + QUtil::hex_encode(std::string(&ch, 1));
 	}
 	else
 	{
@@ -41,6 +40,18 @@ QPDF_Name::unparse()
     return normalizeName(this->name);
 }
 
+QPDFObject::object_type_e
+QPDF_Name::getTypeCode() const
+{
+    return QPDFObject::ot_name;
+}
+
+char const*
+QPDF_Name::getTypeName() const
+{
+    return "name";
+}
+
 std::string
 QPDF_Name::getName() const
 {
